@@ -23,7 +23,10 @@
 #include <libplacebo/filters.h>
 #include <libplacebo/gpu.h>
 #include <libplacebo/shaders/colorspace.h>
+#include <libplacebo/shaders/deinterlacing.h>
+#include <libplacebo/shaders/dithering.h>
 #include <libplacebo/shaders/film_grain.h>
+#include <libplacebo/shaders/icc.h>
 #include <libplacebo/shaders/lut.h>
 #include <libplacebo/shaders/sampling.h>
 #include <libplacebo/shaders/custom.h>
@@ -32,29 +35,54 @@
 PL_API_BEGIN
 
 // Thread-safety: Unsafe
-typedef PL_STRUCT(pl_renderer) *pl_renderer;
+typedef struct pl_renderer_t *pl_renderer;
+
+// Enum values used in pl_renderer_errors_t as a bit positions for error flags
+enum pl_render_error {
+    PL_RENDER_ERR_NONE              = 0,
+    PL_RENDER_ERR_FBO               = 1 << 0,
+    PL_RENDER_ERR_SAMPLING          = 1 << 1,
+    PL_RENDER_ERR_DEBANDING         = 1 << 2,
+    PL_RENDER_ERR_BLENDING          = 1 << 3,
+    PL_RENDER_ERR_OVERLAY           = 1 << 4,
+    PL_RENDER_ERR_PEAK_DETECT       = 1 << 5,
+    PL_RENDER_ERR_FILM_GRAIN        = 1 << 6,
+    PL_RENDER_ERR_FRAME_MIXING      = 1 << 7,
+    PL_RENDER_ERR_DEINTERLACING     = 1 << 8,
+    PL_RENDER_ERR_ERROR_DIFFUSION   = 1 << 9,
+    PL_RENDER_ERR_HOOKS             = 1 << 10,
+    PL_RENDER_ERR_CONTRAST_RECOVERY = 1 << 11,
+};
+
+// Struct describing current renderer state, including internal processing errors,
+// as well as list of signatures of disabled hooks.
+struct pl_render_errors {
+    enum pl_render_error errors;
+    // List containing signatures of disabled hooks
+    const uint64_t *disabled_hooks;
+    int num_disabled_hooks;
+};
 
 // Creates a new renderer object, which is backed by a GPU context. This is a
 // high-level object that takes care of the rendering chain as a whole, from
 // the source textures to the finished frame.
-pl_renderer pl_renderer_create(pl_log log, pl_gpu gpu);
-void pl_renderer_destroy(pl_renderer *rr);
+PL_API pl_renderer pl_renderer_create(pl_log log, pl_gpu gpu);
+PL_API void pl_renderer_destroy(pl_renderer *rr);
 
-// Saves the internal shader cache of this renderer into an abstract cache
-// object that can be saved to disk and later re-loaded to speed up
-// recompilation of shaders. See `pl_dispatch_save` for more information.
-size_t pl_renderer_save(pl_renderer rr, uint8_t *out_cache);
+// Returns current renderer state, see pl_render_errors.
+PL_API struct pl_render_errors pl_renderer_get_errors(pl_renderer rr);
 
-// Load the result of a previous `pl_renderer_save` call. See
-// `pl_dispatch_load` for more information.
-//
-// Note: See the security warnings on `pl_pass_params.cached_program`.
-void pl_renderer_load(pl_renderer rr, const uint8_t *cache);
+// Clears errors state of renderer. If `errors` is NULL, all render errors will
+// be cleared. Otherwise only selected errors/hooks will be cleared.
+// If `PL_RENDER_ERR_HOOKS` is set and `num_disabled_hooks` is 0, clear all hooks.
+// Otherwise only selected hooks will be cleard based on `disabled_hooks` array.
+PL_API void pl_renderer_reset_errors(pl_renderer rr,
+                                     const struct pl_render_errors *errors);
 
 enum pl_lut_type {
     PL_LUT_UNKNOWN = 0,
     PL_LUT_NATIVE,      // applied to raw image contents (after fixing bit depth)
-    PL_LUT_NORMALIZED,  // applied to normalized RGB values
+    PL_LUT_NORMALIZED,  // applied to normalized (HDR) RGB values
     PL_LUT_CONVERSION,  // LUT fully replaces color conversion
 
     // Note: When using a PL_LUT_CONVERSION to replace the YUV->RGB conversion,
@@ -79,12 +107,13 @@ struct pl_render_info {
     const struct pl_dispatch_info *pass;    // information about the shader
     enum pl_render_stage stage;             // the associated render stage
 
-    // For PL_RENDER_STAGE_FRAME, this specifies the chronological index
-    // of this pass within the frame (starting at `index == 0`).
-    //
+    // This specifies the chronological index of this pass within the frame and
+    // stage (starting at `index == 0`).
+    int index;
+
     // For PL_RENDER_STAGE_BLEND, this specifies the number of frames
     // being blended (since that results in a different shader).
-    int index;
+    int count;
 };
 
 // Represents the options used for rendering. These affect the quality of
@@ -92,7 +121,7 @@ struct pl_render_info {
 struct pl_render_params {
     // Configures the algorithms used for upscaling and downscaling,
     // respectively. If left as NULL, then libplacebo will only use inexpensive
-    // sampling (bilinear or neareast neighbour depending on the capabilities
+    // sampling (bilinear or nearest neighbour depending on the capabilities
     // of the hardware / texture).
     //
     // Note: Setting `downscaler` to NULL also implies `skip_anti_aliasing`,
@@ -104,11 +133,17 @@ struct pl_render_params {
     const struct pl_filter_config *upscaler;
     const struct pl_filter_config *downscaler;
 
-    // The number of entries for the scaler LUTs. Defaults to 64 if left unset.
-    int lut_entries;
+    // If set, this overrides the value of `upscaler`/`downscaling` for
+    // subsampled (chroma) planes. These scalers are used whenever the size of
+    // multiple different `pl_plane`s in a single `pl_frame` differ, requiring
+    // adaptation when converting to/from RGB. Note that a value of NULL simply
+    // means "no override". To force built-in scaling explicitly, set this to
+    // `&pl_filter_bilinear`.
+    const struct pl_filter_config *plane_upscaler;
+    const struct pl_filter_config *plane_downscaler;
 
-    // The anti-ringing strength to apply to non-polar filters. See the
-    // equivalent option in `pl_sample_filter_params` for more information.
+    // The anti-ringing strength to apply to filters. See the equivalent option
+    // in `pl_sample_filter_params` for more information.
     float antiringing_strength;
 
     // Configures the algorithm used for frame mixing (when using
@@ -152,9 +187,10 @@ struct pl_render_params {
     // as NULL disables dithering.
     const struct pl_dither_params *dither_params;
 
-    // Configures the settings used to handle ICC profiles, if required. If
-    // NULL, defaults to `&pl_icc_default_params`.
-    const struct pl_icc_params *icc_params;
+    // Configures the error diffusion kernel to use for error diffusion
+    // dithering. If set, this will be used instead of `dither_params` whenever
+    // possible. Leaving this as NULL disables error diffusion.
+    const struct pl_error_diffusion_kernel *error_diffusion;
 
     // Configures the settings used to simulate color blindness, if desired.
     // If NULL, this feature is disabled.
@@ -165,6 +201,26 @@ struct pl_render_params {
     // that the target format has PL_FMT_CAP_BLENDABLE. NULL disables blending.
     const struct pl_blend_params *blend_params;
 
+    // Configures the settings used to deinterlace frames (see
+    // `pl_frame.field`), if required.. If NULL, deinterlacing is "disabled",
+    // meaning interlaced frames are rendered as weaved frames instead.
+    //
+    // Note: As a consequence of how `pl_frame` represents individual fields,
+    // and especially when using the `pl_queue`, this will still result in
+    // frames being redundantly rendered twice. As such, it's highly
+    // recommended to, instead, fully disable deinterlacing by not marking
+    // source frames as interlaced in the first place.
+    const struct pl_deinterlace_params *deinterlace_params;
+
+    // If set, applies an extra distortion matrix to the image, after
+    // scaling and before presenting it to the screen. Can be used for e.g.
+    // fractional rotation.
+    //
+    // Note: The distortion canvas will be set to the size of `target->crop`,
+    // so this cannot effectively draw outside the specified target area,
+    // nor change the aspect ratio of the image.
+    const struct pl_distort_params *distort_params;
+
     // List of custom user shaders / hooks.
     // See <libplacebo/shaders/custom.h> for more information.
     const struct pl_hook * const *hooks;
@@ -178,9 +234,9 @@ struct pl_render_params {
     // are scaled so 1.0 corresponds to the `pl_color_transfer_nominal_peak`.
     //
     // Note: A PL_LUT_CONVERSION fully replaces the color adaptation from
-    // `image` to `target`, including any tone-mapping (if necessary). It has
-    // the same representation as PL_LUT_NATIVE, so in this case the input
-    // and output are (respectively) non-linear light RGB.
+    // `image` to `target`, including any tone-mapping (if necessary) and ICC
+    // profiles. It has the same representation as PL_LUT_NATIVE, so in this
+    // case the input and output are (respectively) non-linear light RGB.
     const struct pl_custom_lut *lut;
     enum pl_lut_type lut_type;
 
@@ -191,6 +247,12 @@ struct pl_render_params {
     float background_transparency; // 0.0 for opaque, 1.0 for fully transparent
     bool skip_target_clearing;
 
+    // If set to a value above 0.0, the output will be rendered with rounded
+    // corners, as if an alpha transparency mask had been applied. The value
+    // indicates the relative fraction of the side length to round - a value
+    // of 1.0 rounds the corners as much as possible.
+    float corner_rounding;
+
     // If true, then transparent images will made opaque by painting them
     // against a checkerboard pattern consisting of alternating colors. If both
     // colors are left as {0}, they default respectively to 93% and 87% gray.
@@ -212,16 +274,6 @@ struct pl_render_params {
     // Significantly speeds up downscaling with high downscaling ratios.
     bool skip_anti_aliasing;
 
-    // Cutoff value for polar sampling. See the equivalent option in
-    // `pl_sample_filter_params` for more information.
-    float polar_cutoff;
-
-    // Allows the peak detection result to be delayed by up to a single frame,
-    // which can sometimes (not always) allow skipping some otherwise redundant
-    // sampling work. Only relevant when peak detection is active (i.e.
-    // params->peak_detect_params is set and the source is HDR).
-    bool allow_delayed_peak_detect;
-
     // Normally, when the size of the `target` used with `pl_render_image_mix`
     // changes, or the render parameters are updated, the internal cache of
     // mixed frames must be discarded in order to re-render all required
@@ -231,16 +283,16 @@ struct pl_render_params {
     // resize, but should make it much more smooth.
     bool preserve_mixing_cache;
 
+    // --- Performance tuning / debugging options
+    // These may affect performance or may make debugging problems easier,
+    // but shouldn't have any effect on the quality.
+
     // Normally, `pl_render_image_mix` will also push single frames through the
     // mixer cache, in order to speed up re-draws. Enabling this option
     // disables that logic, causing single frames to bypass the cache. (Though
     // it will still read from, if they happen to already be cached)
     bool skip_caching_single_frame;
 
-    // --- Performance tuning / debugging options
-    // These may affect performance or may make debugging problems easier,
-    // but shouldn't have any effect on the quality.
-
     // Disables linearization / sigmoidization before scaling. This might be
     // useful when tracking down unexpected image artifacts or excessing
     // ringing, but it shouldn't normally be necessary.
@@ -252,14 +304,8 @@ struct pl_render_params {
     // general-purpose ones.
     bool disable_builtin_scalers;
 
-    // Forces the use of an ICC 3DLUT, even in cases where the use of one is
-    // unnecessary. This is slower, but may improve the quality of the gamut
-    // reduction step, if one is performed.
-    bool force_icc_lut;
-
-    // Ignore ICC profiles attached to either `image` or `target`.
-    // Note: A LUT may still be generated if `force_icc_lut` is also enabled.
-    bool ignore_icc_profiles;
+    // Forces correction of subpixel offsets (using the configured `upscaler`).
+    bool correct_subpixel_offsets;
 
     // Forces the use of dithering, even when rendering to 16-bit FBOs. This is
     // generally pretty pointless because most 16-bit FBOs have high enough
@@ -267,10 +313,18 @@ struct pl_render_params {
     // but this can be used to test the dither code.
     bool force_dither;
 
+    // Disables the gamma-correct dithering logic which normally applies when
+    // dithering to low bit depths. No real use, outside of testing.
+    bool disable_dither_gamma_correction;
+
     // Completely overrides the use of FBOs, as if there were no renderable
     // texture format available. This disables most features.
     bool disable_fbos;
 
+    // Use only low-bit-depth FBOs (8 bits). Note that this also implies
+    // disabling linear scaling and sigmoidization.
+    bool force_low_bit_depth_fbos;
+
     // If this is true, all shaders will be generated as "dynamic" shaders,
     // with any compile-time constants being replaced by runtime-adjustable
     // values. This is generally a performance loss, but has the advantage of
@@ -288,63 +342,42 @@ struct pl_render_params {
     void (*info_callback)(void *priv, const struct pl_render_info *info);
     void *info_priv;
 
-    // --- Deprecated aliases
-    const struct pl_icc_params *lut3d_params PL_DEPRECATED; // fallback for `icc_params`
-    bool force_3dlut PL_DEPRECATED; // fallback for `force_icc_lut`
-
     // --- Deprecated/removed fields
-    bool disable_overlay_sampling PL_DEPRECATED; // no longer used
+    bool allow_delayed_peak_detect PL_DEPRECATED; // moved to pl_peak_detect_params
+    const struct pl_icc_params *icc_params PL_DEPRECATED; // use pl_frame.icc
+    bool ignore_icc_profiles PL_DEPRECATED; // non-functional, just set pl_frame.icc to NULL
+    int lut_entries PL_DEPRECATED; // hard-coded as 256
+    float polar_cutoff PL_DEPRECATED; // hard-coded as 1e-3
 };
 
 // Bare minimum parameters, with no features enabled. This is the fastest
 // possible configuration, and should therefore be fine on any system.
 #define PL_RENDER_DEFAULTS                              \
-    /* set a frame mixer for pl_render_image_mix */     \
-    .frame_mixer        = &pl_filter_oversample,        \
     .color_map_params   = &pl_color_map_default_params, \
-    .lut_entries        = 64,                           \
+    .color_adjustment   = &pl_color_adjustment_neutral, \
     .tile_colors        = {{0.93, 0.93, 0.93},          \
                            {0.87, 0.87, 0.87}},         \
-    .tile_size          = 32,                           \
-    .polar_cutoff       = 0.001,
+    .tile_size          = 32,
 
 #define pl_render_params(...) (&(struct pl_render_params) { PL_RENDER_DEFAULTS __VA_ARGS__ })
-extern const struct pl_render_params pl_render_fast_params;
+PL_API extern const struct pl_render_params pl_render_fast_params;
 
 // This contains the default/recommended options for reasonable image quality,
 // while also not being too terribly slow. All of the *_params structs are
 // defaulted to the corresponding *_default_params, except for deband_params,
-// and peak_detect_params, which are both disabled by default.
+// which is disabled by default.
 //
 // This should be fine on most integrated GPUs, but if it's too slow,
 // consider using `pl_render_fast_params` instead.
-extern const struct pl_render_params pl_render_default_params;
+PL_API extern const struct pl_render_params pl_render_default_params;
 
 // This contains a higher quality preset for better image quality at the cost
 // of quite a bit of performance. In addition to the settings implied by
-// `pl_render_default_params`, it sets the upscaler to `pl_filter_ewa_lanczos`,
-// and enables debanding and peak detection. This should only really be used
-// with a discrete GPU and where maximum image quality is desired.
-extern const struct pl_render_params pl_render_high_quality_params;
-
-// Special filter config for the built-in oversampling algorithm. This is an
-// opaque filter with no meaningful representation. though it has one tunable
-// parameter controlling the threshold at which to switch back to ordinary
-// nearest neighbour sampling. (See `pl_shader_sample_oversample`)
-extern const struct pl_filter_config pl_filter_oversample;
-
-// Backwards compatibility
-#define pl_oversample_frame_mixer pl_filter_oversample
-
-// A list of recommended frame mixer presets, terminated by {0}
-extern const struct pl_filter_preset pl_frame_mixers[];
-extern const int pl_num_frame_mixers; // excluding trailing {0}
-
-// A list of recommended scaler presets, terminated by {0}. This is almost
-// equivalent to `pl_filter_presets` with the exception of including extra
-// built-in filters that don't map to the `pl_filter` architecture.
-extern const struct pl_filter_preset pl_scale_filters[];
-extern const int pl_num_scale_filters; // excluding trailing {0}
+// `pl_render_default_params`, it enables debanding, sets the upscaler to
+// `pl_filter_ewa_lanczossharp`, and uses pl_*_high_quality_params structs where
+// available. This should only really be used with a discrete GPU and where
+// maximum image quality is desired.
+PL_API extern const struct pl_render_params pl_render_high_quality_params;
 
 #define PL_MAX_PLANES 4
 
@@ -360,6 +393,20 @@ struct pl_plane {
     // since the default (PL_TEX_ADDRESS_CLAMP) is very reasonable.
     enum pl_tex_address_mode address_mode;
 
+    // Controls whether or not the `texture` will be considered flipped
+    // vertically with respect to the overall image dimensions. It's generally
+    // preferable to flip planes using this setting instead of the crop in
+    // cases where the flipping is the result of e.g. negative plane strides or
+    // flipped framebuffers (OpenGL).
+    //
+    // Note that any planar padding (due to e.g. size mismatch or misalignment
+    // of subsampled planes) is always at the physical end of the texture
+    // (highest y coordinate) - even if this bool is true. However, any
+    // subsampling shift (`shift_y`) is applied with respect to the flipped
+    // direction. This ensures the correct interpretation when e.g. vertically
+    // flipping 4:2:0 sources by flipping all planes.
+    bool flipped;
+
     // Describes the number and interpretation of the components in this plane.
     // This defines the mapping from component index to the canonical component
     // order (RGBA, YCbCrA or XYZA). It's worth pointing out that this is
@@ -412,9 +459,23 @@ enum pl_overlay_mode {
     PL_OVERLAY_MODE_COUNT,
 };
 
+enum pl_overlay_coords {
+    PL_OVERLAY_COORDS_AUTO = 0,  // equal to SRC/DST_FRAME, respectively
+    PL_OVERLAY_COORDS_SRC_FRAME, // relative to the raw src frame
+    PL_OVERLAY_COORDS_SRC_CROP,  // relative to the src frame crop
+    PL_OVERLAY_COORDS_DST_FRAME, // relative to the raw dst frame
+    PL_OVERLAY_COORDS_DST_CROP,  // relative to the dst frame crop
+    PL_OVERLAY_COORDS_COUNT,
+
+    // Note on rotations: If there is an end-to-end rotation between `src` and
+    // `dst`, then any overlays relative to SRC_FRAME or SRC_CROP will be
+    // rotated alongside the image, while overlays relative to DST_FRAME or
+    // DST_CROP will not.
+};
+
 struct pl_overlay_part {
-    struct pl_rect2df src; // source coordinate with respect to `tex`
-    struct pl_rect2d dst;  // target coordinates with respect to the frame
+    pl_rect2df src; // source coordinate with respect to `pl_overlay.tex`
+    pl_rect2df dst; // target coordinates with respect to `pl_overlay.coords`
 
     // If `mode` is PL_OVERLAY_MONOCHROME, then this specifies the color of
     // this overlay part. The color is multiplied into the sampled texture's
@@ -432,6 +493,9 @@ struct pl_overlay {
     // This controls the coloring mode of this overlay.
     enum pl_overlay_mode mode;
 
+    // Controls which coordinates this overlay is addressed relative to.
+    enum pl_overlay_coords coords;
+
     // This controls the colorspace information for this overlay. The contents
     // of the texture / the value of `color` are interpreted according to this.
     struct pl_color_repr repr;
@@ -440,13 +504,6 @@ struct pl_overlay {
     // The number of parts for this overlay.
     const struct pl_overlay_part *parts;
     int num_parts;
-
-    // (Deprecated) These fields exist for backwards compatibility. They must
-    // not be used as the same times as `tex`. They are interpreted as an
-    // overlay with a single part.
-    struct pl_plane plane PL_DEPRECATED;
-    struct pl_rect2d rect PL_DEPRECATED; // analog to `pl_overlay_part.dst`
-    float base_color[3] PL_DEPRECATED; // analog to `pl_overlay_part.color`
 };
 
 // High-level description of a complete frame, including metadata and planes
@@ -456,11 +513,53 @@ struct pl_frame {
     int num_planes;
     struct pl_plane planes[PL_MAX_PLANES];
 
+    // For interlaced frames. If set, this `pl_frame` corresponds to a single
+    // field of the underlying source textures. `first_field` indicates which
+    // of these fields is ordered first in time. `prev` and `next` should point
+    // to the previous/next frames in the file, or NULL if there are none.
+    //
+    // Note: Setting these fields on the render target has no meaning and will
+    // be ignored.
+    enum pl_field field;
+    enum pl_field first_field;
+    const struct pl_frame *prev, *next;
+
+    // If set, will be called immediately before GPU access to this frame. This
+    // function *may* be used to, for example, perform synchronization with
+    // external APIs (e.g. `pl_vulkan_hold/release`). If your mapping requires
+    // a memcpy of some sort (e.g. pl_tex_transfer), users *should* instead do
+    // the memcpy up-front and avoid the use of these callbacks - because they
+    // might be called multiple times on the same frame.
+    //
+    // This function *may* arbitrarily mutate the `pl_frame`, but it *should*
+    // ideally only update `planes` - in particular, color metadata and so
+    // forth should be provided up-front as best as possible. Note that changes
+    // here will not be reflected back to the structs provided in the original
+    // `pl_render_*` call (e.g. via `pl_frame_mix`).
+    //
+    // Note: Unless dealing with interlaced frames, only one frame will ever be
+    // acquired at a time per `pl_render_*` call. So users *can* safely use
+    // this with, for example, hwdec mappers that can only map a single frame
+    // at a time. When using this with, for example, `pl_render_image_mix`,
+    // each frame to be blended is acquired and release in succession, before
+    // moving on to the next frame. For interlaced frames, the previous and
+    // next frames must also be acquired simultaneously.
+    bool (*acquire)(pl_gpu gpu, struct pl_frame *frame);
+
+    // If set, will be called after a plane is done being used by the GPU,
+    // *including* after any errors (e.g. `acquire` returning false).
+    void (*release)(pl_gpu gpu, struct pl_frame *frame);
+
     // Color representation / encoding / semantics of this frame.
     struct pl_color_repr repr;
     struct pl_color_space color;
 
     // Optional ICC profile associated with this frame.
+    pl_icc_object icc;
+
+    // Alternative to `icc`, this can be used in cases where allocating and
+    // tracking an pl_icc_object externally may be inconvenient. The resulting
+    // profile will be managed internally by the pl_renderer.
     struct pl_icc_profile profile;
 
     // Optional LUT associated with this frame.
@@ -477,7 +576,7 @@ struct pl_frame {
     // Note that `pl_render_image` will map the input crop directly to the
     // output crop, stretching and scaling as needed. If you wish to preserve
     // the aspect ratio, use a dedicated function like pl_rect2df_aspect_copy.
-    struct pl_rect2df crop;
+    pl_rect2df crop;
 
     // Logical rotation of the image, with respect to the underlying planes.
     // For example, if this is PL_ROTATION_90, then the image will be rotated
@@ -497,9 +596,9 @@ struct pl_frame {
     // compensate).
     pl_rotation rotation;
 
-    // A list of additional overlays to render directly on top of this frame.
-    // These overlays will be treated as though they were part of the frame
-    // data, and can be used for things like subtitles or on-screen displays.
+    // A list of additional overlays associated with this frame. Note that will
+    // be rendered directly onto intermediate/cache frames, so changing any of
+    // these overlays may require flushing the renderer cache.
     const struct pl_overlay *overlays;
     int num_overlays;
 
@@ -532,23 +631,23 @@ struct pl_frame {
 // Helper function to infer the chroma location offset for each plane in a
 // frame. This is equivalent to calling `pl_chroma_location_offset` on all
 // subsampled planes' shift_x/shift_y variables.
-void pl_frame_set_chroma_location(struct pl_frame *frame,
-                                  enum pl_chroma_location chroma_loc);
+PL_API void pl_frame_set_chroma_location(struct pl_frame *frame,
+                                         enum pl_chroma_location chroma_loc);
 
 // Fills in a `pl_frame` based on a swapchain frame's FBO and metadata.
-void pl_frame_from_swapchain(struct pl_frame *out_frame,
-                             const struct pl_swapchain_frame *frame);
+PL_API void pl_frame_from_swapchain(struct pl_frame *out_frame,
+                                    const struct pl_swapchain_frame *frame);
 
 // Helper function to determine if a frame is logically cropped or not. In
 // particular, this is useful in determining whether or not an output frame
 // needs to be cleared before rendering or not.
-bool pl_frame_is_cropped(const struct pl_frame *frame);
+PL_API bool pl_frame_is_cropped(const struct pl_frame *frame);
 
 // Helper function to reset a frame to a given RGB color. If the frame's
 // color representation is something other than RGB, the clear color will
 // be adjusted accordingly. `clear_color` should be non-premultiplied.
-void pl_frame_clear_rgba(pl_gpu gpu, const struct pl_frame *frame,
-                         const float clear_color[4]);
+PL_API void pl_frame_clear_rgba(pl_gpu gpu, const struct pl_frame *frame,
+                                const float clear_color[4]);
 
 // Like `pl_frame_clear_rgba` but without an alpha channel.
 static inline void pl_frame_clear(pl_gpu gpu, const struct pl_frame *frame,
@@ -558,6 +657,27 @@ static inline void pl_frame_clear(pl_gpu gpu, const struct pl_frame *frame,
     pl_frame_clear_rgba(gpu, frame, clear_color_rgba);
 }
 
+// Helper functions to return the fixed/inferred pl_frame parameters used
+// for rendering internally. Mutates `image` and `target` in-place to hold
+// the modified values, which are what will actually be used for rendering.
+//
+// This currently includes:
+// - Defaulting all missing pl_color_space/repr parameters
+// - Coalescing all rotation to the target
+// - Rounding and clamping the target crop to pixel boundaries and adjusting the
+//   image crop correspondingly
+//
+// Note: This is idempotent and does not generally alter the effects of a
+// subsequent `pl_render_image` on the same pl_frame pair. (But see the
+// following warning)
+//
+// Warning: This does *not* call pl_frame.acquire/release, and so the returned
+// metadata *may* be incorrect if the acquire callback mutates the pl_frame in
+// nontrivial ways, in particular the crop and color space fields.
+PL_API void pl_frames_infer(pl_renderer rr, struct pl_frame *image,
+                            struct pl_frame *target);
+
+
 // Render a single image to a target using the given parameters. This is
 // fully dynamic, i.e. the params can change at any time. libplacebo will
 // internally detect and flush whatever caches are invalidated as a result of
@@ -576,16 +696,11 @@ static inline void pl_frame_clear(pl_gpu gpu, const struct pl_frame *frame,
 // freely overwritten or discarded by the caller, even the referenced
 // `pl_tex` objects may be freely reused.
 //
-// Note on overlays: `image.overlays` will be rendered directly onto the image,
-// which means they get affected by things like scaling and frame mixing.
-// `target.overlays` will also be rendered, but directly onto the target. They
-// don't even need to be inside `target.crop`.
-//
 // Note: `image` may be NULL, in which case `target.overlays` will still be
 // rendered, but nothing else.
-bool pl_render_image(pl_renderer rr, const struct pl_frame *image,
-                     const struct pl_frame *target,
-                     const struct pl_render_params *params);
+PL_API bool pl_render_image(pl_renderer rr, const struct pl_frame *image,
+                            const struct pl_frame *target,
+                            const struct pl_render_params *params);
 
 // Flushes the internal state of this renderer. This is normally not needed,
 // even if the image parameters, colorspace or target configuration change,
@@ -594,7 +709,13 @@ bool pl_render_image(pl_renderer rr, const struct pl_frame *image,
 // purge some state related to things like HDR peak detection or frame mixing,
 // so calling it is a good idea if the content source is expected to change
 // dramatically (e.g. when switching to a different file).
-void pl_renderer_flush_cache(pl_renderer rr);
+PL_API void pl_renderer_flush_cache(pl_renderer rr);
+
+// Mirrors `pl_get_detected_hdr_metadata`, giving you the current internal peak
+// detection HDR metadata (when peak detection is active). Returns false if no
+// information is available (e.g. not HDR source, peak detection disabled).
+PL_API bool pl_renderer_get_hdr_metadata(pl_renderer rr,
+                                         struct pl_hdr_metadata *metadata);
 
 // Represents a mixture of input frames, distributed temporally.
 //
@@ -664,7 +785,10 @@ struct pl_frame_mix {
     // "next" frames.
 };
 
-// Helper function to calculate the frame mixing radius.
+// Helper function to calculate the base frame mixing radius.
+//
+// Note: When the source FPS exceeds the display FPS, this radius must be
+// increased by the corresponding ratio.
 static inline float pl_frame_mix_radius(const struct pl_render_params *params)
 {
     // For backwards compatibility, allow !frame_mixer->kernel
@@ -674,6 +798,12 @@ static inline float pl_frame_mix_radius(const struct pl_render_params *params)
     return params->frame_mixer->kernel->radius;
 }
 
+// Find closest frame to current PTS by zero-order hold semantics, or NULL.
+PL_API const struct pl_frame *pl_frame_mix_current(const struct pl_frame_mix *mix);
+
+// Find closest frame to current PTS by nearest neighbour semantics, or NULL.
+PL_API const struct pl_frame *pl_frame_mix_nearest(const struct pl_frame_mix *mix);
+
 // Render a mixture of images to the target using the given parameters. This
 // functions much like a generalization of `pl_render_image`, for when the API
 // user has more control over the frame queue / vsync loop, and can provide a
@@ -682,9 +812,35 @@ static inline float pl_frame_mix_radius(const struct pl_render_params *params)
 // This allows libplacebo to perform rudimentary frame mixing / interpolation,
 // in order to eliminate judder artifacts typically associated with
 // source/display frame rate mismatch.
-bool pl_render_image_mix(pl_renderer rr, const struct pl_frame_mix *images,
-                         const struct pl_frame *target,
-                         const struct pl_render_params *params);
+PL_API bool pl_render_image_mix(pl_renderer rr, const struct pl_frame_mix *images,
+                                const struct pl_frame *target,
+                                const struct pl_render_params *params);
+
+// Analog of `pl_frame_infer` corresponding to `pl_render_image_mix`. This
+// function will *not* mutate the frames contained in `mix`, and instead
+// return an adjusted copy of the "reference" frame for that image mix in
+// `out_refimage`, or {0} if the mix is empty.
+PL_API void pl_frames_infer_mix(pl_renderer rr, const struct pl_frame_mix *mix,
+                                struct pl_frame *target, struct pl_frame *out_ref);
+
+// Backwards compatibility with old filters API, may be deprecated.
+// Redundant with pl_filter_configs and masking `allowed` for
+// PL_FILTER_SCALING and PL_FILTER_FRAME_MIXING respectively.
+
+// A list of recommended frame mixer presets, terminated by {0}
+PL_API extern const struct pl_filter_preset pl_frame_mixers[];
+PL_API extern const int pl_num_frame_mixers; // excluding trailing {0}
+
+// A list of recommended scaler presets, terminated by {0}. This is almost
+// equivalent to `pl_filter_presets` with the exception of including extra
+// built-in filters that don't map to the `pl_filter` architecture.
+PL_API extern const struct pl_filter_preset pl_scale_filters[];
+PL_API extern const int pl_num_scale_filters; // excluding trailing {0}
+
+// Deprecated in favor of `pl_cache_save/pl_cache_load` on the `pl_cache`
+// associated with the `pl_gpu` this renderer is using.
+PL_DEPRECATED PL_API size_t pl_renderer_save(pl_renderer rr, uint8_t *out_cache);
+PL_DEPRECATED PL_API void pl_renderer_load(pl_renderer rr, const uint8_t *cache);
 
 PL_API_END
 
