@@ -17,11 +17,15 @@
 
 #ifndef LIBPLACEBO_DAV1D_H_
 #error This header should be included as part of <libplacebo/utils/dav1d.h>
+#elif defined(__cplusplus)
+#error This header cannot be included from C++ define PL_DAV1D_IMPLEMENTATION appropriately
 #else
 
+#include <assert.h>
 #include <stdlib.h>
+#include <string.h>
 
-static inline enum pl_color_system pl_system_from_dav1d(enum Dav1dMatrixCoefficients mc)
+PL_DAV1D_API enum pl_color_system pl_system_from_dav1d(enum Dav1dMatrixCoefficients mc)
 {
     switch (mc) {
         case DAV1D_MC_IDENTITY:     return PL_COLOR_SYSTEM_RGB; // or XYZ (unlikely)
@@ -47,7 +51,7 @@ static inline enum pl_color_system pl_system_from_dav1d(enum Dav1dMatrixCoeffici
     return PL_COLOR_SYSTEM_UNKNOWN;
 }
 
-static inline enum Dav1dMatrixCoefficients pl_system_to_dav1d(enum pl_color_system sys)
+PL_DAV1D_API enum Dav1dMatrixCoefficients pl_system_to_dav1d(enum pl_color_system sys)
 {
     switch (sys) {
     case PL_COLOR_SYSTEM_UNKNOWN:       return DAV1D_MC_UNKNOWN;
@@ -58,6 +62,7 @@ static inline enum Dav1dMatrixCoefficients pl_system_to_dav1d(enum pl_color_syst
     case PL_COLOR_SYSTEM_BT_2020_C:     return DAV1D_MC_BT2020_CL;
     case PL_COLOR_SYSTEM_BT_2100_PQ:    return DAV1D_MC_ICTCP;
     case PL_COLOR_SYSTEM_BT_2100_HLG:   return DAV1D_MC_ICTCP;
+    case PL_COLOR_SYSTEM_DOLBYVISION:   return DAV1D_MC_UNKNOWN; // missing
     case PL_COLOR_SYSTEM_YCGCO:         return DAV1D_MC_SMPTE_YCGCO;
     case PL_COLOR_SYSTEM_RGB:           return DAV1D_MC_IDENTITY;
     case PL_COLOR_SYSTEM_XYZ:           return DAV1D_MC_IDENTITY;
@@ -67,17 +72,17 @@ static inline enum Dav1dMatrixCoefficients pl_system_to_dav1d(enum pl_color_syst
     return DAV1D_MC_UNKNOWN;
 }
 
-static inline enum pl_color_levels pl_levels_from_dav1d(int color_range)
+PL_DAV1D_API enum pl_color_levels pl_levels_from_dav1d(int color_range)
 {
     return color_range ? PL_COLOR_LEVELS_FULL : PL_COLOR_LEVELS_LIMITED;
 }
 
-static inline int pl_levels_to_dav1d(enum pl_color_levels levels)
+PL_DAV1D_API int pl_levels_to_dav1d(enum pl_color_levels levels)
 {
     return levels == PL_COLOR_LEVELS_FULL;
 }
 
-static inline enum pl_color_primaries pl_primaries_from_dav1d(enum Dav1dColorPrimaries prim)
+PL_DAV1D_API enum pl_color_primaries pl_primaries_from_dav1d(enum Dav1dColorPrimaries prim)
 {
     switch (prim) {
     case DAV1D_COLOR_PRI_BT709:         return PL_COLOR_PRIM_BT_709;
@@ -89,7 +94,7 @@ static inline enum pl_color_primaries pl_primaries_from_dav1d(enum Dav1dColorPri
     case DAV1D_COLOR_PRI_SMPTE240:      return PL_COLOR_PRIM_BT_601_525;
     case DAV1D_COLOR_PRI_FILM:          return PL_COLOR_PRIM_FILM_C;
     case DAV1D_COLOR_PRI_BT2020:        return PL_COLOR_PRIM_BT_2020;
-    case DAV1D_COLOR_PRI_XYZ:           return PL_COLOR_PRIM_CIE_1931;
+    case DAV1D_COLOR_PRI_XYZ:           return PL_COLOR_PRIM_UNKNOWN;
     case DAV1D_COLOR_PRI_SMPTE431:      return PL_COLOR_PRIM_DCI_P3;
     case DAV1D_COLOR_PRI_SMPTE432:      return PL_COLOR_PRIM_DISPLAY_P3;
     case DAV1D_COLOR_PRI_EBU3213:       return PL_COLOR_PRIM_EBU_3213;
@@ -98,7 +103,7 @@ static inline enum pl_color_primaries pl_primaries_from_dav1d(enum Dav1dColorPri
     return PL_COLOR_PRIM_UNKNOWN;
 }
 
-static inline enum Dav1dColorPrimaries pl_primaries_to_dav1d(enum pl_color_primaries prim)
+PL_DAV1D_API enum Dav1dColorPrimaries pl_primaries_to_dav1d(enum pl_color_primaries prim)
 {
     switch (prim) {
     case PL_COLOR_PRIM_UNKNOWN:     return DAV1D_COLOR_PRI_UNKNOWN;
@@ -111,19 +116,21 @@ static inline enum Dav1dColorPrimaries pl_primaries_to_dav1d(enum pl_color_prima
     case PL_COLOR_PRIM_APPLE:       return DAV1D_COLOR_PRI_UNKNOWN; // missing
     case PL_COLOR_PRIM_ADOBE:       return DAV1D_COLOR_PRI_UNKNOWN; // missing
     case PL_COLOR_PRIM_PRO_PHOTO:   return DAV1D_COLOR_PRI_UNKNOWN; // missing
-    case PL_COLOR_PRIM_CIE_1931:    return DAV1D_COLOR_PRI_XYZ;
+    case PL_COLOR_PRIM_CIE_1931:    return DAV1D_COLOR_PRI_UNKNOWN; // missing
     case PL_COLOR_PRIM_DCI_P3:      return DAV1D_COLOR_PRI_SMPTE431;
     case PL_COLOR_PRIM_DISPLAY_P3:  return DAV1D_COLOR_PRI_SMPTE432;
     case PL_COLOR_PRIM_V_GAMUT:     return DAV1D_COLOR_PRI_UNKNOWN; // missing
     case PL_COLOR_PRIM_S_GAMUT:     return DAV1D_COLOR_PRI_UNKNOWN; // missing
     case PL_COLOR_PRIM_FILM_C:      return DAV1D_COLOR_PRI_FILM;
+    case PL_COLOR_PRIM_ACES_AP0:    return DAV1D_COLOR_PRI_UNKNOWN; // missing
+    case PL_COLOR_PRIM_ACES_AP1:    return DAV1D_COLOR_PRI_UNKNOWN; // missing
     case PL_COLOR_PRIM_COUNT: abort();
     }
 
     return DAV1D_COLOR_PRI_UNKNOWN;
 }
 
-static inline enum pl_color_transfer pl_transfer_from_dav1d(enum Dav1dTransferCharacteristics trc)
+PL_DAV1D_API enum pl_color_transfer pl_transfer_from_dav1d(enum Dav1dTransferCharacteristics trc)
 {
     switch (trc) {
     case DAV1D_TRC_BT709:           return PL_COLOR_TRC_BT_1886; // EOTF != OETF
@@ -141,7 +148,7 @@ static inline enum pl_color_transfer pl_transfer_from_dav1d(enum Dav1dTransferCh
     case DAV1D_TRC_BT2020_10BIT:    return PL_COLOR_TRC_BT_1886; // EOTF != OETF
     case DAV1D_TRC_BT2020_12BIT:    return PL_COLOR_TRC_BT_1886; // EOTF != OETF
     case DAV1D_TRC_SMPTE2084:       return PL_COLOR_TRC_PQ;
-    case DAV1D_TRC_SMPTE428:        return PL_COLOR_TRC_UNKNOWN; // missing
+    case DAV1D_TRC_SMPTE428:        return PL_COLOR_TRC_ST428;
     case DAV1D_TRC_HLG:             return PL_COLOR_TRC_HLG;
     case DAV1D_TRC_RESERVED: abort();
     }
@@ -149,7 +156,7 @@ static inline enum pl_color_transfer pl_transfer_from_dav1d(enum Dav1dTransferCh
     return PL_COLOR_TRC_UNKNOWN;
 }
 
-static inline enum Dav1dTransferCharacteristics pl_transfer_to_dav1d(enum pl_color_transfer trc)
+PL_DAV1D_API enum Dav1dTransferCharacteristics pl_transfer_to_dav1d(enum pl_color_transfer trc)
 {
     switch (trc) {
     case PL_COLOR_TRC_UNKNOWN:      return DAV1D_TRC_UNKNOWN;
@@ -162,6 +169,7 @@ static inline enum Dav1dTransferCharacteristics pl_transfer_to_dav1d(enum pl_col
     case PL_COLOR_TRC_GAMMA24:      return DAV1D_TRC_UNKNOWN; // missing
     case PL_COLOR_TRC_GAMMA26:      return DAV1D_TRC_UNKNOWN; // missing
     case PL_COLOR_TRC_GAMMA28:      return DAV1D_TRC_BT470BG;
+    case PL_COLOR_TRC_ST428:        return DAV1D_TRC_SMPTE428;
     case PL_COLOR_TRC_PRO_PHOTO:    return DAV1D_TRC_UNKNOWN; // missing
     case PL_COLOR_TRC_PQ:           return DAV1D_TRC_SMPTE2084;
     case PL_COLOR_TRC_HLG:          return DAV1D_TRC_HLG;
@@ -174,7 +182,7 @@ static inline enum Dav1dTransferCharacteristics pl_transfer_to_dav1d(enum pl_col
     return DAV1D_TRC_UNKNOWN;
 }
 
-static inline enum pl_chroma_location pl_chroma_from_dav1d(enum Dav1dChromaSamplePosition loc)
+PL_DAV1D_API enum pl_chroma_location pl_chroma_from_dav1d(enum Dav1dChromaSamplePosition loc)
 {
     switch (loc) {
     case DAV1D_CHR_UNKNOWN:     return PL_CHROMA_UNKNOWN;
@@ -185,7 +193,7 @@ static inline enum pl_chroma_location pl_chroma_from_dav1d(enum Dav1dChromaSampl
     return PL_CHROMA_UNKNOWN;
 }
 
-static inline enum Dav1dChromaSamplePosition pl_chroma_to_dav1d(enum pl_chroma_location loc)
+PL_DAV1D_API enum Dav1dChromaSamplePosition pl_chroma_to_dav1d(enum pl_chroma_location loc)
 {
     switch (loc) {
     case PL_CHROMA_UNKNOWN:         return DAV1D_CHR_UNKNOWN;
@@ -219,8 +227,8 @@ static inline float pl_fixed0_16(uint16_t n)
 // Align to a power of 2
 #define PL_ALIGN2(x, align) (((x) + (align) - 1) & ~((align) - 1))
 
-static inline void pl_frame_from_dav1dpicture(struct pl_frame *out,
-                                              const Dav1dPicture *picture)
+PL_DAV1D_API void pl_frame_from_dav1dpicture(struct pl_frame *out,
+                                             const Dav1dPicture *picture)
 {
     const Dav1dSequenceHeader *seq_hdr = picture->seq_hdr;
     int num_planes;
@@ -257,7 +265,6 @@ static inline void pl_frame_from_dav1dpicture(struct pl_frame *out,
         .color = {
             .primaries = pl_primaries_from_dav1d(seq_hdr->pri),
             .transfer = pl_transfer_from_dav1d(seq_hdr->trc),
-            .light = PL_COLOR_LIGHT_UNKNOWN,
         },
         .repr = {
             .sys = pl_system_from_dav1d(seq_hdr->mtrx),
@@ -292,39 +299,48 @@ static inline void pl_frame_from_dav1dpicture(struct pl_frame *out,
 
     const Dav1dContentLightLevel *cll = picture->content_light;
     if (cll) {
-        out->color.sig_peak = cll->max_content_light_level / PL_COLOR_SDR_WHITE;
-        out->color.sig_avg = cll->max_frame_average_light_level / PL_COLOR_SDR_WHITE;
+        out->color.hdr.max_cll = cll->max_content_light_level;
+        out->color.hdr.max_fall = cll->max_frame_average_light_level;
     }
 
     // This overrides the CLL values above, if both are present
     const Dav1dMasteringDisplay *md = picture->mastering_display;
     if (md) {
-        out->color.sig_peak = pl_fixed24_8(md->max_luminance) / PL_COLOR_SDR_WHITE;
-        out->color.sig_floor = pl_fixed18_14(md->min_luminance) / PL_COLOR_SDR_WHITE;
+        out->color.hdr.max_luma = pl_fixed24_8(md->max_luminance);
+        out->color.hdr.min_luma = pl_fixed18_14(md->min_luminance);
+        out->color.hdr.prim = (struct pl_raw_primaries) {
+            .red.x   = pl_fixed0_16(md->primaries[0][0]),
+            .red.y   = pl_fixed0_16(md->primaries[0][1]),
+            .green.x = pl_fixed0_16(md->primaries[1][0]),
+            .green.y = pl_fixed0_16(md->primaries[1][1]),
+            .blue.x  = pl_fixed0_16(md->primaries[2][0]),
+            .blue.y  = pl_fixed0_16(md->primaries[2][1]),
+            .white.x = pl_fixed0_16(md->white_point[0]),
+            .white.y = pl_fixed0_16(md->white_point[1]),
+        };
     }
 
-    // Make sure this value is more or less legal
-    if (out->color.sig_peak < 1.0 || out->color.sig_peak > 50.0)
-        out->color.sig_peak = 0.0;
-
     if (picture->frame_hdr->film_grain.present) {
         const Dav1dFilmGrainData *fg = &picture->frame_hdr->film_grain.data;
-        struct pl_av1_grain_data *av1 = &out->av1_grain;
-        *av1 = (struct pl_av1_grain_data) {
-            .grain_seed = fg->seed,
-            .num_points_y = fg->num_y_points,
-            .chroma_scaling_from_luma = fg->chroma_scaling_from_luma,
-            .num_points_uv = { fg->num_uv_points[0], fg->num_uv_points[1] },
-            .scaling_shift = fg->scaling_shift,
-            .ar_coeff_lag = fg->ar_coeff_lag,
-            .ar_coeff_shift = (int) fg->ar_coeff_shift,
-            .grain_scale_shift = fg->grain_scale_shift,
-            .uv_mult = { fg->uv_mult[0], fg->uv_mult[1] },
-            .uv_mult_luma = { fg->uv_luma_mult[0], fg->uv_luma_mult[1] },
-            .uv_offset = { fg->uv_offset[0], fg->uv_offset[1] },
-            .overlap = fg->overlap_flag,
+        out->film_grain = (struct pl_film_grain_data) {
+            .type = PL_FILM_GRAIN_AV1,
+            .seed = fg->seed,
+            .params.av1 = {
+                .num_points_y = fg->num_y_points,
+                .chroma_scaling_from_luma = fg->chroma_scaling_from_luma,
+                .num_points_uv = { fg->num_uv_points[0], fg->num_uv_points[1] },
+                .scaling_shift = fg->scaling_shift,
+                .ar_coeff_lag = fg->ar_coeff_lag,
+                .ar_coeff_shift = (int) fg->ar_coeff_shift,
+                .grain_scale_shift = fg->grain_scale_shift,
+                .uv_mult = { fg->uv_mult[0], fg->uv_mult[1] },
+                .uv_mult_luma = { fg->uv_luma_mult[0], fg->uv_luma_mult[1] },
+                .uv_offset = { fg->uv_offset[0], fg->uv_offset[1] },
+                .overlap = fg->overlap_flag,
+            },
         };
 
+        struct pl_av1_grain_data *av1 = &out->film_grain.params.av1;
         memcpy(av1->points_y, fg->y_points, sizeof(av1->points_y));
         memcpy(av1->points_uv, fg->uv_points, sizeof(av1->points_uv));
         memcpy(av1->ar_coeffs_y, fg->ar_coeffs_y, sizeof(av1->ar_coeffs_y));
@@ -344,8 +360,8 @@ static inline void pl_frame_from_dav1dpicture(struct pl_frame *out,
     }
 }
 
-static void pl_swapchain_colors_from_dav1dpicture(struct pl_swapchain_colors *out_colors,
-                                                  const Dav1dPicture *picture)
+PL_DAV1D_API void pl_swapchain_colors_from_dav1dpicture(struct pl_swapchain_colors *out_colors,
+                                                            const Dav1dPicture *picture)
 {
     struct pl_frame frame;
     pl_frame_from_dav1dpicture(&frame, picture);
@@ -399,15 +415,15 @@ static void pl_dav1dpicture_unref(void *priv)
     }
 }
 
-static inline bool pl_upload_dav1dpicture(pl_gpu gpu,
-                                          struct pl_frame *out,
-                                          pl_tex tex[3],
-                                          const struct pl_dav1d_upload_params *params)
+PL_DAV1D_API bool pl_upload_dav1dpicture(pl_gpu gpu,
+                                             struct pl_frame *out,
+                                             pl_tex tex[3],
+                                             const struct pl_dav1d_upload_params *params)
 {
     Dav1dPicture *pic = params->picture;
     pl_frame_from_dav1dpicture(out, pic);
     if (!params->film_grain)
-        out->av1_grain = (struct pl_av1_grain_data) {0};
+        out->film_grain.type = PL_FILM_GRAIN_NONE;
 
     const int bytes = (pic->p.bpc + 7) / 8; // rounded up
     int sub_x = 0, sub_y = 0;
@@ -430,7 +446,6 @@ static inline bool pl_upload_dav1dpicture(pl_gpu gpu,
             .width          = pic->p.w,
             .height         = pic->p.h,
             .pixel_stride   = bytes,
-            .row_stride     = pic->stride[0],
             .component_size = {bytes * 8},
             .component_map  = {0},
         }, {
@@ -439,7 +454,6 @@ static inline bool pl_upload_dav1dpicture(pl_gpu gpu,
             .width          = pic->p.w >> sub_x,
             .height         = pic->p.h >> sub_y,
             .pixel_stride   = bytes,
-            .row_stride     = pic->stride[1],
             .component_size = {bytes * 8},
             .component_map  = {1},
         }, {
@@ -448,7 +462,6 @@ static inline bool pl_upload_dav1dpicture(pl_gpu gpu,
             .width          = pic->p.w >> sub_x,
             .height         = pic->p.h >> sub_y,
             .pixel_stride   = bytes,
-            .row_stride     = pic->stride[1],
             .component_size = {bytes * 8},
             .component_map  = {2},
         },
@@ -471,15 +484,23 @@ static inline bool pl_upload_dav1dpicture(pl_gpu gpu,
     }
 
     for (int p = 0; p < out->num_planes; p++) {
-        if (buf) {
-            data[p].buf = buf;
-            data[p].buf_offset = (uintptr_t) pic->data[p] - (uintptr_t) buf->data;
+        ptrdiff_t stride = p > 0 ? pic->stride[1] : pic->stride[0];
+        if (stride < 0) {
+            data[p].pixels = (uint8_t *) pic->data[p] + stride * (data[p].height - 1);
+            data[p].row_stride = -stride;
+            out->planes[p].flipped = true;
         } else {
             data[p].pixels = pic->data[p];
-            if (ref) {
-                data[p].priv = ref;
-                data[p].callback = pl_dav1dpicture_unref;
-            }
+            data[p].row_stride = stride;
+        }
+
+        if (buf) {
+            data[p].buf = buf;
+            data[p].buf_offset = (uintptr_t) data[p].pixels - (uintptr_t) buf->data;
+            data[p].pixels = NULL;
+        } else if (ref) {
+            data[p].priv = ref;
+            data[p].callback = pl_dav1dpicture_unref;
         }
 
         if (!pl_upload_plane(gpu, &out->planes[p], &tex[p], &data[p])) {
@@ -499,11 +520,14 @@ static inline bool pl_upload_dav1dpicture(pl_gpu gpu,
     return true;
 }
 
-static inline int pl_allocate_dav1dpicture(Dav1dPicture *p, void *cookie)
+PL_DAV1D_API int pl_allocate_dav1dpicture(Dav1dPicture *p, void *cookie)
 {
     pl_gpu gpu = cookie;
-    if (!gpu->limits.max_mapped_size || !gpu->limits.buf_transfer)
+    if (!gpu->limits.max_mapped_size || !gpu->limits.host_cached ||
+        !gpu->limits.buf_transfer)
+    {
         return DAV1D_ERR(ENOTSUP);
+    }
 
     // Copied from dav1d_default_picture_alloc
     const int hbd = p->p.bpc > 8;
@@ -516,10 +540,10 @@ static inline int pl_allocate_dav1dpicture(Dav1dPicture *p, void *cookie)
     p->stride[1] = has_chroma ? (aligned_w >> ss_hor) << hbd : 0;
 
     // Align strides up to multiples of the GPU performance hints
-    p->stride[0] = PL_ALIGN2(p->stride[0], gpu->limits.align_tex_xfer_stride);
-    p->stride[1] = PL_ALIGN2(p->stride[1], gpu->limits.align_tex_xfer_stride);
+    p->stride[0] = PL_ALIGN2(p->stride[0], gpu->limits.align_tex_xfer_pitch);
+    p->stride[1] = PL_ALIGN2(p->stride[1], gpu->limits.align_tex_xfer_pitch);
 
-    // Aligning offsets to 4 also implicitly aligns to the texel size (1 or 2)
+    // Aligning offsets to 4 also implicitly aligns to the texel alignment (1 or 2)
     size_t off_align = PL_ALIGN2(gpu->limits.align_tex_xfer_offset, 4);
     const size_t y_sz = PL_ALIGN2(p->stride[0] * aligned_h, off_align);
     const size_t uv_sz = PL_ALIGN2(p->stride[1] * (aligned_h >> ss_ver), off_align);
@@ -533,11 +557,11 @@ static inline int pl_allocate_dav1dpicture(Dav1dPicture *p, void *cookie)
     if (total_size > gpu->limits.max_mapped_size)
         return DAV1D_ERR(ENOMEM);
 
-    pl_buf buf = pl_buf_create(gpu, &(struct pl_buf_params) {
+    pl_buf buf = pl_buf_create(gpu, pl_buf_params(
         .size = total_size,
         .host_mapped = true,
         .memory_type = PL_BUF_MEM_HOST,
-    });
+    ));
 
     if (!buf)
         return DAV1D_ERR(ENOMEM);
@@ -567,16 +591,15 @@ static inline int pl_allocate_dav1dpicture(Dav1dPicture *p, void *cookie)
     return 0;
 }
 
-static inline void pl_release_dav1dpicture(Dav1dPicture *p, void *cookie)
+PL_DAV1D_API void pl_release_dav1dpicture(Dav1dPicture *p, void *cookie)
 {
-    pl_gpu gpu = cookie;
     struct pl_dav1dalloc *alloc = p->allocator_data;
     if (!alloc)
         return;
 
     assert(alloc->magic[0] == PL_MAGIC0);
     assert(alloc->magic[1] == PL_MAGIC1);
-    assert(alloc->gpu == gpu);
+    assert(alloc->gpu == cookie);
     pl_buf_destroy(alloc->gpu, &alloc->buf);
     free(alloc);
 
